#include <stdio.h>
#include <stdlib.h>
/**
U lekciji o binarnim stablima smo pomenuli da njihova struktura zavisi od redosljeda dodavanja elemenata u stablo
i da se moze desiti da se stablo navije u jednu stranu, tj. da ne bude izbalansirano. U krajnjem slucaju stablo moze izgledati kao lista.
Na taj nacin se gube sve prednosti stabala, a to su efikasnije operacije trazenja elemenata, dodavanja i brisanja elementa,
koje bi trebalo da su slozenosti manje od O(n), gdje je n broj cvorova u stablu.

AVL stablo je samobalansirajuce u kojem se garantuje da ce operacije dodavanja i brisanja i trazenja elementa biti slozenosti O(logn).
Po definiciji, binarno stablo se smatra AVL stablom ukoliko je balansni faktor svakog cvora u stablu vrijednost iz skupa {-1, 0, 1}.
U binarnom stablu, balansni faktor je vrijednost koja se vezuje za cvor i predstavlja razliku u visinama izmedju njegovog lijevog i desnog podstabla.
Dakle, u AVL stablu je razlika visina lijevog i desnog podstabla za svaki cvor po apsolutnoj vrijednosti manja ili jednaka 1.

Da bi se stablo odrzavalo balansiranim, poslije svake operacije dodavanja ili brisanja, u cvorovima ciji balansni faktor je poremecen (<-1 ili >1)
potrebno je izvrsiti odgovarajuce rotacije. Postoje 4 razicita scenarija za rotaciju i njih mozete vidjeti u kodu ispod ili u objasnjenju.

Predavanje o AVL stablima mozete naci na ovom linku: https://www.youtube.com/watch?v=jDM6_TnYIqE
**/
struct cvor{
    int kljuc;
    struct cvor* lijevi;
    struct cvor* desni;
    int visina;
};

// funkcija koja vraca visinu stabla, ciji je korijen cvor c
int visina(struct cvor* c){
    if(c==NULL)
        return 0;

    return c->visina;
}
// funkcija koja racuna razliku izmedju visine lijevog i desnog podstabla
// koristicemo je da odredimo da li i u koju stranu treba stablo rotirati da bi se izbalansiralo
int balans(struct cvor* c){
    if(c==NULL)
        return 0;
    return visina(c->lijevi) - visina(c->desni);
}

// procedura koja azurira visinu stabla sa korijenom u cvoru c
// poziva se nakon rotacije
void azurirajVisinu(struct cvor* c){
    c->visina = visina(c->lijevi) > visina(c->desni) ? visina(c->lijevi)+1 : visina(c->desni) + 1;
}

// funkcija koja vrsi rotaciju u desno oko datog cvora c i vraca adresu cvora koji dolazi na njegovo mjesto
// da bi funkcija bila jasna, pozeljno je nacrtati sliku
struct cvor* rotacijaUDesno(struct cvor* c){
    printf("Rotacija u desno oko %d\n", c->kljuc);
    struct cvor* temp = c->lijevi; // ovaj cvor zapamtimo jer ce on da dodje na mjesto cvora c nakon rotacije i njega treba vratiti

    c->lijevi = temp->desni;
    temp->desni = c;

    azurirajVisinu(c);
    azurirajVisinu(temp);

    return temp;
}

struct cvor* rotacijaULijevo(struct cvor* c){
    printf("Rotacija u lijevo oko %d\n", c->kljuc);
    struct cvor* temp = c->desni;

    c->desni = temp->lijevi;
    temp->lijevi = c;

    azurirajVisinu(c);
    azurirajVisinu(temp);

    return temp;
}

// funkcija koja dodaje novi cvor u AVL stablo i vraca novi korijen stabla
// izvor: https://www.geeksforgeeks.org/avl-tree-set-1-insertion/
/***Mozete probati nerekurzivnu implementaciju.***/
struct cvor* dodaj(struct cvor* korijen, int novi_kljuc){
    // prvi dio algoritma dodavanja novog elementa u AVL stablo je identican binarnom stablu pretrazivanja
    if(korijen==NULL){
        struct cvor* novi = malloc(sizeof(struct cvor));
        novi->kljuc = novi_kljuc;
        novi->lijevi = NULL;
        novi->desni = NULL;
        novi->visina = 1;
        return novi; // kako je stablo bilo prazno, ovaj cvor je novi korijen stabla
    }else if(novi_kljuc < korijen->kljuc){
        korijen->lijevi = dodaj(korijen->lijevi, novi_kljuc);
    }else if(novi_kljuc > korijen->kljuc){
        korijen->desni = dodaj(korijen->desni, novi_kljuc);
    }else{
        return;
    }

    azurirajVisinu(korijen);
    int b = balans(korijen);
    // ako korijen vise nije izbalansiran tj. ako je |b|>1, moramo ga balansirati
    // tada razlikujemo 4 scenarija

    // 1. ako je stablo nagnuto u lijevu stranu, a novi cvor je u lijevom podstablu lijevog sina, vrsimo rotaciju u desno oko korijena
    if(b>1 && novi_kljuc < korijen->lijevi->kljuc)
        return rotacijaUDesno(korijen);

    // 2. ako je stablo nagnuto u desnu stranu, a novi cvor je u desnom podstablu lijevog sina, vrsimo rotaciju u lijevo oko korijena
    if(b<-1 && novi_kljuc > korijen->desni->kljuc)
        return rotacijaULijevo(korijen);

    /* 3. ako je stablo nagnuto u lijevu stranu, a novi cvor je u desnom podstablu lijevog sina
          tada vrsimo rotaciju u lijevo oko lijevog sina korijena, pa rotaciju u desno oko korijena da bi izbalansirali stablo*/
    if(b>1 && novi_kljuc > korijen->lijevi->kljuc){
        korijen->lijevi = rotacijaULijevo(korijen->lijevi);
        return rotacijaUDesno(korijen);
    }

    /* 4. ako je stablo nagnuto u desnu stranu, a novi cvor je u ljevom podstablu desnog sina
          tada vrsimo rotaciju u desno oko desnog sina korijena, pa rotaciju u lijevo oko korijena da bi izbalansirali stablo*/
    if(b<-1 && novi_kljuc < korijen->desni->kljuc){
        korijen->desni = rotacijaUDesno(korijen->desni);
        return rotacijaULijevo(korijen);
    }

    return korijen;
}

// preorder-om je najlakse provjeriti da li je dobra struktura AVL stabla
void preorder(struct cvor* korijen){
    if(korijen==NULL)
        return;

    printf("kljuc: %d visina: %d\n", korijen->kljuc, korijen->visina);
    preorder(korijen->lijevi);
    preorder(korijen->desni);
}

/**ZA VJEZBU: Implementirati operaciju brisanja.***/
/***
Ova operacija je prosirenje operacije brisanja iz binarnog stabla pretrazivanja.
Nakon brisanja kao iz nebalansiranog binarnog stabla pretrazivanja, mogu se desiti 4 scenarija, u kojima je potrebno vrsiti odredjene rotacije u stablu.

Implementaciju pokusajte da odradite samostalno, nakon toga pogledajte objasnjenje na https://www.geeksforgeeks.org/avl-tree-set-2-deletion/.
Na kraju, na ovom linku imate i implementaciju ove operacije u programskom jeziku C.
**/
int main()
{
    struct cvor* korijen = NULL;

    korijen = dodaj(korijen, 20);
    korijen = dodaj(korijen, 27);
    korijen = dodaj(korijen, 17);
    korijen = dodaj(korijen, 33);
    korijen = dodaj(korijen, 25);
    korijen = dodaj(korijen, 39);
    korijen = dodaj(korijen, 6);
    korijen = dodaj(korijen, 18);
    korijen = dodaj(korijen, 19);

    printf("\n");
    preorder(korijen);
    return 0;
}
